import logging
from typing import Any

from aiomisc import PeriodicCallback, Service


log = logging.getLogger(__name__)


class PeriodicService(Service):

    __required__ = ("interval",)

    interval = None  # type: float # in seconds
    delay = 0  # type: float # in seconds

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        self.periodic = PeriodicCallback(self.callback)

    async def start(self) -> None:
        self.periodic.start(self.interval, delay=self.delay, loop=self.loop)
        log.info("Periodic service %s started", self)

    async def stop(self, err: Exception = None) -> None:
        if self.periodic.task:
            await self.periodic.task
        self.periodic.stop()
        log.info("Periodic service %s is stopped", self)

    async def callback(self) -> Any:
        raise NotImplementedError

    def __str__(self) -> str:
        return "{}(interval={},delay={})".format(
            self.__class__.__name__,
            self.interval,
            self.delay,
        )



import logging
import typing as t
from enum import Enum, IntEnum, unique


@unique
class LogFormat(IntEnum):
    stream = 0
    color = 1
    json = 2
    syslog = 3

    @classmethod
    def choices(cls) -> t.Tuple[str, ...]:
        return tuple(cls._member_names_)    # type: ignore


class LogLevel(IntEnum):
    critical = logging.CRITICAL
    error = logging.ERROR
    warning = logging.WARNING
    info = logging.INFO
    debug = logging.DEBUG
    notset = logging.NOTSET

    @classmethod
    def choices(cls) -> t.Tuple[str, ...]:
        return tuple(cls._member_names_)    # type: ignore


class DateFormat(Enum):
    color = "%Y-%m-%d %H:%M:%S"
    stream = "[%Y-%m-%d %H:%M:%S]"

    # Optimization: special value ``%s`` date will
    # not formatted just returns record created time
    json = "%s"
    syslog = None



import asyncio
import logging
import logging.handlers
import os
import sys
import typing as t

from .enum import LogFormat, LogLevel
from .formatter import color_formatter, json_handler
from .wrap import wrap_logging_handler


LOG_LEVEL: t.Optional[t.Any] = None
LOG_FORMAT: t.Optional[t.Any] = None

try:
    import contextvars
    LOG_LEVEL = contextvars.ContextVar("LOG_LEVEL", default=logging.INFO)
    LOG_FORMAT = contextvars.ContextVar("LOG_FORMAT", default=LogFormat.color)
except ImportError:
    pass


DEFAULT_FORMAT = "%(levelname)s:%(name)s:%(message)s"


def create_logging_handler(
    log_format: LogFormat = LogFormat.color,
    date_format: str = None, **kwargs: t.Any
) -> logging.Handler:

    if LOG_FORMAT is not None:
        LOG_FORMAT.set(log_format)

    if log_format == LogFormat.stream:
        handler = logging.StreamHandler()   # type: logging.Handler
        if date_format and date_format is not Ellipsis:
            formatter = logging.Formatter(
                "%(asctime)s " + DEFAULT_FORMAT, datefmt=date_format,
            )
        else:
            formatter = logging.Formatter(DEFAULT_FORMAT)

        handler.setFormatter(formatter)
        return handler
    elif log_format == LogFormat.json:
        return json_handler(date_format=date_format, **kwargs)
    elif log_format == LogFormat.color:
        return color_formatter(date_format=date_format, **kwargs)
    elif log_format == LogFormat.syslog:
        if date_format:
            sys.stderr.write("Can not apply \"date_format\" for syslog\n")
            sys.stderr.flush()

        formatter = logging.Formatter("%(message)s")

        if os.path.exists("/dev/log"):
            handler = logging.handlers.SysLogHandler(address="/dev/log")
        else:
            handler = logging.handlers.SysLogHandler()

        handler.setFormatter(formatter)
        return handler

    raise NotImplementedError


def basic_config(
    level: t.Union[int, str] = logging.INFO,
    log_format: t.Union[str, LogFormat] = LogFormat.color,
    buffered: bool = True, buffer_size: int = 1024,
    flush_interval: t.Union[int, float] = 0.2,
    loop: asyncio.AbstractEventLoop = None,
    **kwargs: t.Any
) -> None:

    if isinstance(level, str):
        level = LogLevel[level]

    logging.basicConfig()
    logger = logging.getLogger()
    logger.handlers.clear()

    if isinstance(log_format, str):
        log_format = LogFormat[log_format]

    handler = create_logging_handler(log_format, **kwargs)

    if buffered:
        handler = wrap_logging_handler(
            handler,
            buffer_size=buffer_size,
            flush_interval=flush_interval,
            loop=loop,
        )

    if LOG_LEVEL is not None:
        LOG_LEVEL.set(level)

    # noinspection PyArgumentList
    logging.basicConfig(
        level=int(level),
        handlers=[handler],
    )



import asyncio
import itertools
import logging.handlers
import socket
from functools import wraps
from multiprocessing import cpu_count
from typing import (
    Any, Awaitable, Callable, Iterable, List, Optional, Tuple, TypeVar, Union,
)

from .thread_pool import ThreadPoolExecutor


T = TypeVar("T")
TimeoutType = Union[int, float]

try:
    import uvloop  # type: ignore
    event_loop_policy = uvloop.EventLoopPolicy()
except ImportError:
    event_loop_policy = asyncio.DefaultEventLoopPolicy()


log = logging.getLogger(__name__)


def chunk_list(iterable: Iterable[T], size: int) -> Iterable[List[T]]:
    """
    Split list or generator by chunks with fixed maximum size.
    """

    iterable = iter(iterable)

    item = list(itertools.islice(iterable, size))
    while item:
        yield item
        item = list(itertools.islice(iterable, size))


OptionsType = Iterable[Tuple[int, int, int]]


def bind_socket(
    *args: Any, address: str, port: int, options: OptionsType = (),
    reuse_addr: bool = True, reuse_port: bool = False,
    proto_name: Optional[str] = None
) -> socket.socket:
    """

    Bind socket and set ``setblocking(False)`` for just created socket.
    This detects ``address`` format and select socket family automatically.

    :param args: which will be passed to stdlib's socket constructor (optional)
    :param address: bind address
    :param port: bind port
    :param options: Tuple of pairs which contain socket option
                    to set and the option value.
    :param reuse_addr: set socket.SO_REUSEADDR
    :param reuse_port: set socket.SO_REUSEPORT
    :param proto_name: protocol name which will be logged after binding
    :return: socket.socket
    """

    if not args:
        if ":" in address:
            args = (socket.AF_INET6, socket.SOCK_STREAM)
        else:
            args = (socket.AF_INET, socket.SOCK_STREAM)

    sock = socket.socket(*args)
    sock.setblocking(False)

    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, int(reuse_addr))
    if hasattr(socket, "SO_REUSEPORT"):
        sock.setsockopt(
            socket.SOL_SOCKET, socket.SO_REUSEPORT, int(reuse_port),
        )
    else:
        log.warning("SO_REUSEPORT is not implemented by underlying library.")

    for level, option, value in options:
        sock.setsockopt(level, option, value)

    unix_address_family = getattr(socket, "AF_UNIX", None)
    if sock.family == unix_address_family:
        proto_name = proto_name or "unix"
        sock.bind(address)
    else:
        proto_name = proto_name or "tcp"
        sock.bind((address, port))

    sock_addr = sock.getsockname()
    if not isinstance(sock_addr, str):
        sock_addr = sock_addr[:2]

    if sock.family == socket.AF_INET6:
        log.info("Listening %s://[%s]:%s", proto_name, *sock_addr)
    elif sock.family == unix_address_family:
        log.info("Listening %s://%s", proto_name, sock_addr)
    else:
        log.info("Listening %s://%s:%s", proto_name, *sock_addr)

    return sock


def create_default_event_loop(
    pool_size: Optional[int] = None,
    policy: asyncio.AbstractEventLoopPolicy = event_loop_policy,
    debug: bool = False,
) -> Tuple[asyncio.AbstractEventLoop, ThreadPoolExecutor]:
    """
    Creates an event loop and thread pool executor

    :param pool_size: thread pool maximal size
    :param policy: event loop policy
    :param debug: set ``loop.set_debug(True)`` if True
    """
    try:
        asyncio.get_event_loop().close()
    except RuntimeError:
        pass  # event loop is not created yet

    asyncio.set_event_loop_policy(policy)

    loop = asyncio.new_event_loop()
    loop.set_debug(debug)
    asyncio.set_event_loop(loop)

    pool_size = pool_size or cpu_count()
    thread_pool = ThreadPoolExecutor(pool_size, statistic_name="default")
    loop.set_default_executor(thread_pool)

    return loop, thread_pool


def new_event_loop(
    pool_size: Optional[int] = None,
    policy: asyncio.AbstractEventLoopPolicy = event_loop_policy,
) -> asyncio.AbstractEventLoop:
    loop, thread_pool = create_default_event_loop(pool_size, policy)
    return loop


def shield(func: Callable[..., Awaitable[T]]) -> Callable[..., Awaitable[T]]:
    """
    Simple and useful decorator for wrap the coroutine to `asyncio.shield`.

    >>> @shield
    ... async def non_cancelable_func():
    ...     await asyncio.sleep(1)

    """

    async def awaiter(future: Awaitable[Any]) -> Any:
        return await future

    @wraps(func)
    def wrap(*args: Any, **kwargs: Any) -> Awaitable[Any]:
        return wraps(func)(awaiter)(asyncio.shield(func(*args, **kwargs)))

    return wrap


class SelectResult:
    def __init__(self, length: int):
        self.length = length
        self.result_idx = None      # type: Optional[int]
        self.is_exception = None    # type: Optional[bool]
        self.value = None           # type: Any

    def set_result(self, idx: int, value: Any, is_exception: bool) -> None:
        if self.result_idx is not None:
            return

        self.value = value
        self.result_idx = idx
        self.is_exception = is_exception

    def result(self) -> T:
        if self.is_exception:
            raise self.value
        return self.value

    def done(self) -> bool:
        return self.result_idx is not None

    def __iter__(self) -> Iterable[Optional[T]]:
        for i in range(self.length):
            if i == self.result_idx:
                yield self.value
            else:
                yield None


def cancel_tasks(tasks: Iterable[asyncio.Future]) -> asyncio.Future:
    """
    All passed tasks will be cancelled and a new task will be returned.

    :param tasks: tasks which will be cancelled
    """

    future = asyncio.get_event_loop().create_future()
    future.set_result(None)

    if not tasks:
        return future

    cancelled_tasks = []
    exc = asyncio.CancelledError()

    for task in tasks:
        if task.done():
            continue

        if isinstance(task, asyncio.Task):
            task.cancel()
            cancelled_tasks.append(task)

        elif isinstance(task, asyncio.Future):
            task.set_exception(exc)

        else:
            log.warning(
                "Skipping object %r because it's not a Task or Future", task,
            )

    if not cancelled_tasks:
        return future

    waiter = asyncio.ensure_future(
        asyncio.gather(
            *cancelled_tasks, return_exceptions=True
        ),
    )

    return waiter


async def _select_waiter(
    idx: int, awaitable: Awaitable[T],
    result: SelectResult,
) -> None:
    try:
        ret = await awaitable
    except asyncio.CancelledError:
        raise
    except Exception as e:
        return result.set_result(idx, e, is_exception=True)

    result.set_result(idx, ret, is_exception=False)


async def select(
    *awaitables: Awaitable[Any],
    return_exceptions: bool = False, cancel: bool = True,
    timeout: Optional[TimeoutType] = None,
    wait: bool = True, loop: Optional[asyncio.AbstractEventLoop] = None
) -> SelectResult:
    """

    :param awaitables: awaitable objects
    :param return_exceptions: if True exception will not be raised
                              just returned as result
    :param cancel: cancel unfinished coroutines (default True)
    :param timeout: execution timeout
    :param wait: when False and ``cancel=True``, unfinished coroutines will
                 be cancelled in the background.
    :param loop: event loop
    """

    loop = loop or asyncio.get_event_loop()
    result = SelectResult(len(awaitables))

    coroutines = [
        loop.create_task(_select_waiter(idx, coroutine, result))
        for idx, coroutine in enumerate(awaitables)
    ]

    _, pending = await loop.create_task(
        asyncio.wait(
            coroutines,
            timeout=timeout,
            return_when=asyncio.FIRST_COMPLETED,
        ),
    )

    if cancel:
        cancelling = cancel_tasks(pending)

        if wait:
            await cancelling

    if result.is_exception and not return_exceptions:
        result.result()

    return result


def awaitable(
    func: Callable[..., Union[T, Awaitable[T]]],
) -> Callable[..., Awaitable[T]]:
    """

    Decorator wraps function and returns a function which returns
    awaitable object. In case than a function returns a future,
    the original future will be returned. In case then the function
    returns a coroutine, the original coroutine will be returned.
    In case than function returns non-awaitable object, it's will
    be wrapped to a new coroutine which just returns this object.
    It's useful when you don't want to check function result before
    use it in ``await`` expression.
    """

    # Avoid python 3.8+ warning
    if asyncio.iscoroutinefunction(func):
        return func     # type: ignore

    async def awaiter(obj: T) -> T:
        return obj

    @wraps(func)
    def wrap(*args: Any, **kwargs: Any) -> Awaitable[T]:
        result = func(*args, **kwargs)

        if hasattr(result, "__await__"):
            return result       # type: ignore
        if asyncio.iscoroutine(result) or asyncio.isfuture(result):
            return result       # type: ignore

        return awaiter(result)  # type: ignore

    return wrap



import asyncio
import logging
import pickle
import signal
import socket
import sys
from os import urandom
from types import FrameType
from typing import Any, Tuple, Union, Optional

from aiomisc.log import basic_config
from aiomisc.worker_pool.constants import (
    HASHER, INET_AF, SALT_SIZE, SIGNAL, Header, PacketTypes,
)


def on_signal(signum: int, frame: FrameType) -> None:
    raise asyncio.CancelledError


def main() -> Optional[int]:
    address: Union[str, Tuple[str, int]]
    cookie: bytes
    identity: str

    (
        address, cookie, identity, log_level, log_format,
    ) = pickle.load(sys.stdin.buffer)

    basic_config(
        level=log_level,
        log_format=log_format,
        buffered=False,
    )

    family = (
        socket.AF_UNIX if isinstance(address, str) else INET_AF
    )

    with socket.socket(family, socket.SOCK_STREAM) as sock:
        logging.debug("Connecting...")
        try:
            sock.connect(address)
        except ConnectionRefusedError:
            logging.error("Failed to establish IPC.")
            return 2

        def send(packet_type: PacketTypes, data: Any) -> None:
            payload = pickle.dumps(data)
            sock.send(Header.pack(packet_type.value, len(payload)))
            sock.send(payload)

        def receive() -> Tuple[PacketTypes, Any]:
            header = sock.recv(Header.size)

            if not header:
                raise ValueError("No data")

            packet_type, payload_length = Header.unpack(header)
            payload = sock.recv(payload_length)
            return PacketTypes(packet_type), pickle.loads(payload)

        def auth(cookie: bytes) -> None:
            hasher = HASHER()
            salt = urandom(SALT_SIZE)
            send(PacketTypes.AUTH_SALT, salt)
            hasher.update(salt)
            hasher.update(cookie)
            send(PacketTypes.AUTH_DIGEST, hasher.digest())

            packet_type, value = receive()
            if packet_type == PacketTypes.AUTH_OK:
                return value

            raise RuntimeError(PacketTypes(packet_type), value)

        def step() -> bool:
            try:
                packet_type, (func, args, kwargs) = receive()
            except ConnectionResetError:
                logging.error("Pool connection closed")
                return False
            except ValueError:
                return False

            if packet_type == packet_type.REQUEST:
                response_type = PacketTypes.RESULT
                try:
                    result = func(*args, **kwargs)
                except asyncio.CancelledError:
                    logging.exception("Request cancelled for %r", func)
                    send(PacketTypes.CANCELLED, asyncio.CancelledError)
                    return True
                except Exception as e:
                    response_type = PacketTypes.EXCEPTION
                    result = e
                    logging.exception("Exception when processing request")

                send(response_type, result)
            return True

        logging.debug("Starting authorization")

        try:
            auth(cookie)
        except ConnectionResetError:
            logging.error("Failed to authorize process in pool. Exiting")
            return 3

        del cookie

        send(PacketTypes.IDENTITY, identity)
        logging.debug("Worker ready")

        signal.signal(SIGNAL, on_signal)

        try:
            while step():
                pass
            return 0
        except KeyboardInterrupt:
            return 1


if __name__ == "__main__":
    rc = main()
    exit(rc or 0)



import asyncio
import typing as t

from ..context import Context, get_context
from ..utils import cancel_tasks


class ServiceMeta(type):
    def __new__(
        cls, name: str, bases: t.Tuple, namespace: t.Dict, **kwds: t.Any
    ) -> t.Any:
        instance = type.__new__(
            cls, name, bases, dict(namespace),
        )

        for key in ("__async_required__", "__required__"):
            setattr(instance, key, frozenset(getattr(instance, key, ())))

        check_instance = all(
            asyncio.iscoroutinefunction(getattr(instance, method))
            for method in instance.__async_required__   # type: ignore
        )

        if not check_instance:
            raise TypeError(
                "Following methods must be coroutine functions",
                tuple(
                    "%s.%s" % (name, m)
                    for m in instance.__async_required__    # type: ignore
                ),
            )

        return instance


class Service(metaclass=ServiceMeta):
    __async_required__ = "start", "stop"    # type: t.Tuple[str, ...]
    __required__ = ()                       # type: t.Tuple[str, ...]

    def __init__(self, **kwargs: t.Any):
        lost_kw = self.__required__ - kwargs.keys()
        if lost_kw:
            raise AttributeError("Absent attributes", lost_kw)

        self._set_params(**kwargs)
        self.__context = None           # type: t.Optional[Context]
        self.__start_event = None       # type: t.Optional[asyncio.Event]

    @property
    def start_event(self) -> asyncio.Event:
        if self.__start_event is None:
            raise RuntimeError

        return self.__start_event

    @property
    def context(self) -> Context:
        if self.__context is None:
            self.__context = get_context()
        return self.__context

    def set_loop(self, loop: asyncio.AbstractEventLoop) -> None:
        # noinspection PyAttributeOutsideInit
        self.loop = loop
        self.__start_event = asyncio.Event()

    def _set_params(self, **kwargs: t.Any) -> None:
        for name, value in kwargs.items():
            setattr(self, name, value)

    async def start(self) -> t.Any:
        raise NotImplementedError

    async def stop(self, exception: Exception = None) -> t.Any:
        pass


class SimpleServer(Service):
    def __init__(self, **kwargs: t.Any):
        self.server = None      # type: t.Optional[asyncio.AbstractServer]
        self.tasks = set()      # type: t.Set[asyncio.Task]
        super().__init__(**kwargs)

    def create_task(self, coro: t.Awaitable[t.Any]) -> asyncio.Task:
        task = self.loop.create_task(coro)
        self.tasks.add(task)
        task.add_done_callback(self.tasks.remove)
        return task

    async def start(self) -> None:
        raise NotImplementedError

    async def stop(self, exc: Exception = None) -> None:
        await cancel_tasks(self.tasks)

        if self.server:
            self.server.close()



import asyncio
import logging
import logging.handlers
import time
from contextlib import suppress
from typing import Optional, Union

from ..thread_pool import run_in_new_thread


def _thread_flusher(
    handler: logging.handlers.MemoryHandler,
    flush_interval: Union[float, int],
    loop: asyncio.AbstractEventLoop,
) -> None:
    def has_no_target() -> bool:
        return True

    def has_target() -> bool:
        return bool(handler.target)

    is_target = has_no_target

    if isinstance(handler, logging.handlers.MemoryHandler):
        is_target = has_target

    while not loop.is_closed() and is_target():
        with suppress(Exception):
            if handler.buffer:
                handler.flush()

        time.sleep(flush_interval)


def wrap_logging_handler(
    handler: logging.Handler,
    loop: Optional[asyncio.AbstractEventLoop] = None,
    buffer_size: int = 1024,
    flush_interval: Union[float, int] = 0.1,
) -> logging.Handler:
    loop = loop or asyncio.get_event_loop()

    buffered_handler = logging.handlers.MemoryHandler(
        buffer_size,
        target=handler,
        flushLevel=logging.CRITICAL,
    )

    run_in_new_thread(
        _thread_flusher, args=(
            buffered_handler, flush_interval, loop,
        ), no_return=True, statistic_name="logger",
    )

    return buffered_handler



import asyncio
import logging
import typing as t
from functools import partial

from . import utils
from .counters import Statistic


log = logging.getLogger(__name__)
ExceptionsType = t.Tuple[t.Type[Exception], ...]
CallbackType = t.Callable[..., t.Union[t.Awaitable[t.Any], t.Any]]


class PeriodicCallbackStatistic(Statistic):
    call_count: int
    done: int
    fail: int
    sum_time: float


class PeriodicCallback:
    """
    .. note::

        When the periodic function executes longer then execution interval a
        next call will be skipping and warning will be logged.

    """

    __slots__ = (
        "_cb", "_closed", "_task", "_loop", "_handle", "__name",
        "_statistic",
    )

    def __init__(
        self, coroutine_func: CallbackType,
        *args: t.Any, **kwargs: t.Any
    ):

        self._statistic = PeriodicCallbackStatistic()
        self.__name = repr(coroutine_func)
        self._cb = partial(
            utils.awaitable(coroutine_func), *args, **kwargs
        )
        self._closed = False    # type: t.Optional[bool]
        self._handle = None     # type: t.Optional[asyncio.Handle]
        self._task = None       # type: t.Optional[asyncio.Future]

    async def _run(self, suppress_exceptions: ExceptionsType = ()) -> None:
        try:
            await self._cb()
        except suppress_exceptions:
            return
        except asyncio.CancelledError:
            raise
        except Exception:
            log.exception("Periodic task error:")

    def start(
        self, interval: t.Union[int, float],
        loop: asyncio.AbstractEventLoop = None, *,
        delay: t.Union[int, float] = 0,
        shield: bool = False,
        suppress_exceptions: ExceptionsType = ()
    ) -> None:
        if self._task and not self._task.done():
            raise asyncio.InvalidStateError

        current_loop = loop or asyncio.get_event_loop()
        # noinspection PyAttributeOutsideInit
        self._loop = current_loop   # type: asyncio.AbstractEventLoop

        self._closed = False

        def periodic() -> None:
            if self._loop.is_closed():
                return

            if self._task and not self._task.done():
                log.warning("Task %r still running skipping", self)
                return

            del self._task
            self._task = None

            if self._closed:
                return

            runner = utils.shield(self._run) if shield else self._run
            self._task = self._loop.create_task(
                runner(suppress_exceptions),        # type: ignore
            )

            start_time = self._loop.time()

            self._task.add_done_callback(call)
            self._task.add_done_callback(lambda t: do_stat(t, start_time))

        def do_stat(task: asyncio.Task, start_time: float) -> None:
            self._statistic.call_count += 1
            self._statistic.sum_time += self._loop.time() - start_time

            if task.cancelled():
                self._statistic.fail += 1
            elif task.exception():
                self._statistic.fail += 1
            else:
                self._statistic.done += 1

        def call(*_: t.Any) -> None:
            if self._handle is not None:
                self._handle.cancel()
                del self._handle

            self._handle = self._loop.call_later(
                interval, periodic,
            )

        self._loop.call_later(delay, self._loop.call_soon_threadsafe, periodic)

    def stop(self) -> asyncio.Task:
        self._closed = True

        if self._task is None:
            self._task = self._loop.create_future()
            self._task.set_exception(
                RuntimeError("Callback not started"),
            )
        elif not self._task.done():
            self._task.cancel()

        if self._handle:
            self._handle.cancel()

        return self.task

    def __repr__(self) -> str:
        return "%s(%s)" % (self.__class__.__name__, self.__name)

    @property
    def task(self) -> asyncio.Task:
        return self._task   # type: ignore
